/// <reference types="node" />
import { Bucket as GoogleBucket } from '@google-cloud/storage';
import { S3 } from 'aws-sdk';
import { Readable, Writable } from 'stream';
import { Driver, ListCloudFilesOptions } from './driver';
import { BucketFile, BucketFileDeleted, BucketType, ListArg, ListOptions, ListResult } from './types';
export interface BucketOptions {
    driver: Driver;
    log: boolean;
}
export declare function newBucket(opts: BucketOptions): BucketImpl<unknown>;
export interface Bucket {
    type: BucketType;
    name: string;
    readonly s3?: S3;
    readonly googleBucket?: GoogleBucket;
    exists(path: string): Promise<boolean>;
    /**
     * Get and return a BucketFile for a given path (will do a cloud bucket query).
     * Returns null if not found. Throw exception if any other exception than notfound.
     */
    getFile(path: String): Promise<BucketFile | null>;
    list(optsOrPrefix?: ListArg): Promise<ListResult>;
    listFiles(optsOrPrefix?: ListArg): Promise<BucketFile[]>;
    /**
     * Will copy one or more file to a destination file (then require single match, i.e. no glob),
     * or to a destination folder.
     * @param prefixOrGlob full file name, or prefix or glob. If multiple match, to need dest must be a dir (end with '/')
     * @param dest can be dir path  (copying multiple file and preserving filename and relative dir structure), or full file name (require single match)
     */
    copy(prefixOrGlob: string, dest: string | BucketFile): Promise<void>;
    /**
     * Download one or more remote bucket file to a local file or a folder structure.
     * @param prefixOrGlob
     * @param localDir
     *  If end with '/' then all files from the prefixOrGlob will be downloaded with their originial filename (and relative folder structure).
     *  Otherwise, if full file name, then, make sure there is onyl one matching bucket source, and copy to this file destination (to rename on download)
     */
    download(prefixOrGlob: string, localPath: string): Promise<BucketFile[]>;
    downloadAsText(path: string): Promise<string>;
    upload(localPath: string, path: string): Promise<BucketFile[]>;
    uploadContent(path: string, content: string): Promise<void>;
    createReadStream(path: string): Promise<Readable>;
    createWriteStream(path: string): Promise<Writable>;
    /**
     * Delete a single file.
     * @param path
     */
    delete(path: string): Promise<boolean>;
    deleteAll(files: BucketFile[]): Promise<BucketFileDeleted[]>;
}
declare class BucketImpl<F> implements Bucket {
    driver: Driver<F>;
    log: boolean;
    constructor(opts: BucketOptions);
    get type(): BucketType;
    get name(): string;
    get s3(): S3 | undefined;
    get googleBucket(): GoogleBucket | undefined;
    toFile(cf: F): BucketFile;
    exists(path: string): Promise<boolean>;
    /**
     * Get and return a BucketFile for a given path (will do a cloud bucket query).
     * Returns null if not found. Throw exception if any other exception than notfound.
     */
    getFile(path: String): Promise<BucketFile | null>;
    listFiles(optsOrPrefix?: ListArg): Promise<BucketFile[]>;
    list(optsOrPrefix?: ListArg): Promise<ListResult>;
    /**
     * Will copy one or more file to a destination file (then require single match, i.e. no glob),
     * or to a destination folder.
     * @param prefixOrGlob full file name, or prefix or glob. If multiple match, to need dest must be a dir (end with '/')
     * @param dest can be dir path  (copying multiple file and preserving filename and relative dir structure), or full file name (require single match)
     */
    copy(prefixOrGlob: string, dest: string | BucketFile): Promise<void>;
    /**
     * Download one or more remote bucket file to a local file or a folder structure.
     * @param prefixOrGlob
     * @param localDir
     *  If end with '/' then all files from the prefixOrGlob will be downloaded with their originial filename (and relative folder structure).
     *  Otherwise, if full file name, then, make sure there is onyl one matching bucket source, and copy to this file destination (to rename on download)
     */
    download(prefixOrGlob: string, localPath: string): Promise<BucketFile[]>;
    downloadAsText(path: string): Promise<string>;
    upload(localPath: string, remotePath: string): Promise<BucketFile[]>;
    uploadContent(path: string, content: string): Promise<void>;
    createReadStream(path: string): Promise<Readable>;
    createWriteStream(path: string): Promise<Writable>;
    /**
     * Delete a single file.
     * @param path
     */
    delete(path: string): Promise<boolean>;
    deleteAll(files: BucketFile[]): Promise<BucketFileDeleted[]>;
}
/**
 * Build the full destination path from the local path name and the destPath
 * - If `destPath` ends with `/`, then baseName of `localPath` is concatenated.
 * - Otherwise, `destPath` is the fullDestPath.
 *
 * @throws exception if destPath is not present.
 */
export declare function buildFullDestPath(localPath: string, destPath: string): string;
export declare function getContentType(path: string): string | undefined;
export declare function parseListOptions(optsOrPrefix?: ListOptions | string): ListCloudFilesOptions;
/**
 * Return a clean prefix and glob when defined in the string. Clean prefix, meaning, glob less one,
 * that can be passed to most cloud storage api.
 *
 * @param prefixOrGlob undefined, null, e.g., 'some-prefix', 'folder/', 'folder/glob-pattern.*'
 * @returns {prefix, glob, baseDir}
 * 					- prefix is the first characters until the first glob character ('*')
 * 					- glob is prefixOrGlob value if it is a glob, otherwise undefined.
 * 					- baseDir is the eventual longest directory path without any glob char (ending with '/')
 */
export declare function parsePrefixOrGlob(prefixOrGlob?: string): {
    prefix: string | undefined;
    glob: string | undefined;
    baseDir: string | undefined;
};
export {};
