"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parsePrefixOrGlob = exports.parseListOptions = exports.getContentType = exports.buildFullDestPath = exports.newBucket = void 0;
const fs_extra_plus_1 = require("fs-extra-plus");
const mime_types_1 = require("mime-types");
const Path = __importStar(require("path"));
function newBucket(opts) {
    return new BucketImpl(opts);
}
exports.newBucket = newBucket;
class BucketImpl {
    constructor(opts) {
        this.driver = opts.driver;
        this.log = opts.log;
    }
    get type() {
        return this.driver.type;
    }
    get name() {
        return this.driver.name;
    }
    get s3() {
        if (this.driver.type === 's3') {
            return this.driver.s3;
        }
    }
    get googleBucket() {
        if (this.driver.type === 'gs') {
            return this.driver.googleBucket;
        }
    }
    toFile(cf) {
        const bucketFile = this.driver.toFile(cf);
        bucketFile.bucket = this; // driver does not know/add bucket property
        return bucketFile;
    }
    async exists(path) {
        return this.driver.exists(path);
    }
    /**
     * Get and return a BucketFile for a given path (will do a cloud bucket query).
     * Returns null if not found. Throw exception if any other exception than notfound.
     */
    async getFile(path) {
        const cloudFile = await this.driver.getCloudFile(path);
        return (cloudFile == null) ? null : this.toFile(cloudFile);
    }
    async listFiles(optsOrPrefix) {
        return (await this.list(optsOrPrefix)).files;
    }
    async list(optsOrPrefix) {
        const cloudFilesOptions = parseListOptions(optsOrPrefix);
        const cloudFilesResult = await this.driver.listCloudFiles(cloudFilesOptions);
        const { dirs, nextMarker } = cloudFilesResult;
        const files = cloudFilesResult.files.map(cf => this.toFile(cf));
        return { files, dirs, nextMarker };
    }
    /**
     * Will copy one or more file to a destination file (then require single match, i.e. no glob),
     * or to a destination folder.
     * @param prefixOrGlob full file name, or prefix or glob. If multiple match, to need dest must be a dir (end with '/')
     * @param dest can be dir path  (copying multiple file and preserving filename and relative dir structure), or full file name (require single match)
     */
    async copy(prefixOrGlob, dest) {
        // return this.driver.copy(prefixOrGlob, dest);
        const cloudFiles = (await this.driver.listCloudFiles(parseListOptions(prefixOrGlob))).files;
        const destBucket = (typeof dest === 'string') ? this : dest.bucket;
        const destPath = (typeof dest === 'string') ? dest : dest.path;
        const isDestPathDir = destPath.endsWith('/');
        // If not a local directory, make sure we have only one file.
        // TODO: might check if the pathOrGlob is a glob as well to prevent it (in case there is only one match)
        if (!isDestPathDir && cloudFiles.length > 1) {
            throw new Error(`Cannot copy multiple files ${prefixOrGlob} to the same bucket file ${destPath}. Download to a directory (end with '/') to download multipel file.`);
        }
        const { baseDir } = parsePrefixOrGlob(prefixOrGlob);
        const files = [];
        for (let cf of cloudFiles) {
            const remotePath = this.driver.getPath(cf);
            const destFilePath = (isDestPathDir) ? getDestPath(baseDir, remotePath, destPath) : destPath;
            if (this.log) {
                process.stdout.write(`Copying ${this.type}://${this.name}/${remotePath} to ${destBucket.type}://${destBucket.name}/${destFilePath}`);
            }
            try {
                await this.driver.copyCloudFile(cf, { bucket: destBucket, path: destFilePath });
                if (this.log) {
                    process.stdout.write(` - DONE\n`);
                }
            }
            catch (ex) {
                if (this.log) {
                    process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
                }
                throw ex;
            }
        }
        // return files;
    }
    /**
     * Download one or more remote bucket file to a local file or a folder structure.
     * @param prefixOrGlob
     * @param localDir
     *  If end with '/' then all files from the prefixOrGlob will be downloaded with their originial filename (and relative folder structure).
     *  Otherwise, if full file name, then, make sure there is onyl one matching bucket source, and copy to this file destination (to rename on download)
     */
    async download(prefixOrGlob, localPath) {
        const isLocalPathDir = localPath.endsWith('/');
        const cloudFiles = (await this.driver.listCloudFiles(parseListOptions(prefixOrGlob))).files;
        // If not a local directory, make sure we have only one file.
        // TODO: might check if the pathOrGlob is a glob as well to prevent it (in case there is only one match)
        if (!isLocalPathDir && cloudFiles.length > 1) {
            throw new Error(`Cannot copy multiple files ${prefixOrGlob} to the same local file ${localPath}. Download to a directory (end with '/') to download multipel file.`);
        }
        const files = [];
        const { baseDir } = parsePrefixOrGlob(prefixOrGlob);
        for (let cf of cloudFiles) {
            const remotePath = this.driver.getPath(cf);
            const localFilePath = (isLocalPathDir) ? getDestPath(baseDir, remotePath, localPath) : localPath;
            const localPathDir = Path.dirname(localFilePath);
            await fs_extra_plus_1.mkdirp(localPathDir);
            if (this.log) {
                process.stdout.write(`Downloading ${this.type}://${this.name}/${remotePath} to ${localFilePath}`);
            }
            try {
                await this.driver.downloadCloudFile(cf, localFilePath);
                if (this.log) {
                    process.stdout.write(` - DONE\n`);
                }
                const file = { bucket: this, path: remotePath, size: -1, local: localFilePath };
                files.push(file);
            }
            catch (ex) {
                if (this.log) {
                    process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
                }
                throw ex;
            }
        }
        return files;
    }
    downloadAsText(path) {
        return this.driver.downloadAsText(path);
    }
    async upload(localPath, remotePath) {
        const bucketFiles = [];
        console.log('->> upload', localPath);
        if (localPath.endsWith('/')) {
            localPath = localPath + '**/*.*';
        }
        const isLocalGlob = localPath.includes('*');
        const { baseDir } = parsePrefixOrGlob(localPath);
        const localFiles = await fs_extra_plus_1.glob(localPath);
        for (const localPath of localFiles) {
            // if we have an localFileExpression (globs), then, we build the fullDestPath relative to the baseDir of the glob (last / before the first *)
            const fullDestPath = (isLocalGlob) ? getDestPath(baseDir, localPath, remotePath) : buildFullDestPath(localPath, remotePath);
            const contentType = getContentType(localPath);
            if (this.log) {
                process.stdout.write(`Uploading file ${localPath} to ${this.type}://${this.name}/${fullDestPath}`);
            }
            try {
                const cloudFile = await this.driver.uploadCloudFile(localPath, fullDestPath, contentType);
                const bucketFile = this.toFile(cloudFile);
                bucketFiles.push(bucketFile);
                if (this.log) {
                    process.stdout.write(` - DONE\n`);
                }
            }
            catch (ex) {
                if (this.log) {
                    process.stdout.write(` - FAIL - ABORT - Cause: ${ex}\n`);
                }
                throw ex;
            }
        }
        return bucketFiles;
    }
    uploadContent(path, content) {
        const contentType = getContentType(path);
        return this.driver.uploadCloudContent(path, content, contentType);
    }
    createReadStream(path) {
        return this.driver.createReadStream(path);
    }
    createWriteStream(path) {
        const contentType = getContentType(path);
        return this.driver.createWriteStream(path, contentType);
    }
    /**
     * Delete a single file.
     * @param path
     */
    async delete(path) {
        let deleted = false;
        if (!path) {
            throw new Error(`ERROR - Can't delete null or empty path`);
        }
        try {
            if (this.log) {
                process.stdout.write(`Deleting ${this.type}://${this.name}/${path}`);
            }
            deleted = await this.driver.deleteCloudFile(path);
            if (this.log) {
                process.stdout.write(` - DONE\n`);
            }
        }
        catch (ex) {
            throw new Error(`ERROR - cloud-bucket - Cannot delete ${path} for bucket ${this.name}. Cause: ${ex}`);
        }
        return deleted;
    }
    async deleteAll(files) {
        const filesInfo = [];
        // validate that all files are same bucket
        for (const file of files) {
            // check if same bucket
            if (file.bucket !== this) {
                throw new Error(`Cannot delete file from another bucket ${this.name} should match file bucket ${file.bucket.name}`);
            }
        }
        for (const file of files) {
            const deleted = await this.driver.deleteCloudFile(file.path);
            filesInfo.push({ ...file, deleted });
        }
        return filesInfo;
    }
}
//#region    ---------- Utils ---------- 
function getDestPath(baseDir, remotePath, destPathDir) {
    const baseName = Path.basename(remotePath);
    const filePath = (baseDir) ? Path.relative(baseDir, remotePath) : baseName;
    const destPath = `${destPathDir}${filePath}`;
    return destPath;
}
/**
 * Build the full destination path from the local path name and the destPath
 * - If `destPath` ends with `/`, then baseName of `localPath` is concatenated.
 * - Otherwise, `destPath` is the fullDestPath.
 *
 * @throws exception if destPath is not present.
 */
function buildFullDestPath(localPath, destPath) {
    // we do not have a dest path, throw error
    if (!destPath) {
        throw new Error('No depthPath');
    }
    let fullDestPath;
    // if it is a folder, we just concatinate the base name
    if (destPath.endsWith('/')) {
        const srcBaseName = Path.basename(localPath);
        fullDestPath = destPath + srcBaseName;
    }
    // if the destPath is not a folder, assume it is the new file name.
    else {
        fullDestPath = destPath;
    }
    return fullDestPath;
}
exports.buildFullDestPath = buildFullDestPath;
function getContentType(path) {
    let ct = mime_types_1.lookup(path);
    let contentType = (ct) ? ct : undefined;
    return contentType;
}
exports.getContentType = getContentType;
function parseListOptions(optsOrPrefix) {
    if (optsOrPrefix == null) {
        optsOrPrefix = ''; // for now, default
    }
    const opts = (typeof optsOrPrefix === 'string') ? { prefix: optsOrPrefix } : optsOrPrefix;
    const { prefix, glob } = parsePrefixOrGlob(opts.prefix);
    const { directory, limit, marker } = opts;
    return { prefix, glob, directory, limit, marker };
}
exports.parseListOptions = parseListOptions;
/**
 * Return a clean prefix and glob when defined in the string. Clean prefix, meaning, glob less one,
 * that can be passed to most cloud storage api.
 *
 * @param prefixOrGlob undefined, null, e.g., 'some-prefix', 'folder/', 'folder/glob-pattern.*'
 * @returns {prefix, glob, baseDir}
 * 					- prefix is the first characters until the first glob character ('*')
 * 					- glob is prefixOrGlob value if it is a glob, otherwise undefined.
 * 					- baseDir is the eventual longest directory path without any glob char (ending with '/')
 */
function parsePrefixOrGlob(prefixOrGlob) {
    let glob;
    let prefix;
    let baseDir;
    if (prefixOrGlob && prefixOrGlob.length > 0) {
        const firstWildIdx = prefixOrGlob.indexOf('*');
        // if it has a '*' then it is a pattern
        if (firstWildIdx > 0) {
            glob = prefixOrGlob;
            prefix = prefixOrGlob.substring(0, firstWildIdx);
        }
        // otherwise, it is just a 
        else {
            prefix = prefixOrGlob;
        }
    }
    if (prefix) {
        const lastSlashIdx = prefix.lastIndexOf('/');
        if (lastSlashIdx > -1) {
            baseDir = prefix.substring(0, lastSlashIdx + 1);
        }
    }
    return { prefix, glob, baseDir };
}
exports.parsePrefixOrGlob = parsePrefixOrGlob;
//#endregion ---------- /Utils ----------
