"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.S3Driver = exports.getS3Driver = void 0;
const aws_sdk_1 = require("aws-sdk");
const fs_extra_plus_1 = require("fs-extra-plus");
const stream_1 = require("stream");
const micromatch = require("micromatch");
async function getS3Driver(cfg) {
    const credentials = new aws_sdk_1.Credentials(cfg.access_key_id, cfg.access_key_secret);
    // Create S3 service object
    const s3 = new aws_sdk_1.S3({ apiVersion: '2006-03-01', credentials });
    return new S3Driver(s3, cfg.bucketName);
}
exports.getS3Driver = getS3Driver;
/**
 * Custom Writable to trigger finish/close event manually on upload
 * TODO: Needs to check if this create some side effect.
 */
class S3UploadWriteStream extends stream_1.PassThrough {
    emit(event) {
        if (event !== 'finish' && event !== 'close') {
            super.emit(event);
            return true;
        }
        else {
            return false;
        }
    }
    triggerFinishAndClose() {
        super.emit('finish');
        super.emit('close');
    }
}
class S3Driver {
    constructor(s3, bucketName) {
        this.s3 = s3;
        this.baseParams = { Bucket: bucketName };
    }
    get type() {
        return 's3';
    }
    get name() {
        return this.baseParams.Bucket;
    }
    toFile(awsFile) {
        if (!awsFile) {
            throw new Error(`No awsFile`);
        }
        const updated = (awsFile.LastModified) ? awsFile.LastModified.toISOString() : undefined;
        return {
            path: awsFile.Key,
            size: awsFile.Size,
            contentType: awsFile.ContentType,
            updated: updated
        };
    }
    getPath(obj) {
        return obj.Key; // TODO: need to investigate when Key is empty in S3. 
    }
    async exists(path) {
        const file = await this.getCloudFile(path);
        return (file) ? true : false;
    }
    async getCloudFile(path) {
        try {
            const object = await this.s3.headObject({ ...this.baseParams, ...{ Key: path } }).promise();
            // bucket: this,
            // 	path,
            // 	updated,
            // 	size: object.ContentLength,
            // 		contentType: object.ContentType	
            const { ContentLength, ContentType, LastModified, ETag } = object;
            const Key = path;
            const Size = ContentLength;
            const awsFile = { Key, Size, LastModified, ETag, ContentType };
            return awsFile;
        }
        catch (ex) {
            //  if NotFound, return false
            if (ex.code === 'NotFound') {
                return null;
            }
            // otherwise, propagate the exception
            else {
                throw ex;
            }
        }
    }
    async listCloudFiles(opts) {
        var _a;
        const { prefix, glob, directory, limit, marker } = opts;
        // build the list params
        let listParams = {};
        if (prefix) {
            listParams.Prefix = prefix;
        }
        if (directory) {
            listParams.Delimiter = '/';
        }
        if (limit != null) {
            listParams.MaxKeys = limit;
        }
        if (marker != null) {
            listParams.Marker = marker;
        }
        const params = { ...this.baseParams, ...listParams };
        // perform the s3 list request
        try {
            const awsResult = await this.s3.listObjects(params).promise();
            const awsFiles = awsResult.Contents;
            // if glob, filter again the result
            let files = (!glob) ? awsFiles : awsFiles.filter(af => micromatch.isMatch(af.Key, glob));
            let dirs = undefined;
            if (directory && awsResult.CommonPrefixes) {
                // Note: for now, match the gcp driver, undefined if empty
                const prefixes = (_a = awsResult.CommonPrefixes) === null || _a === void 0 ? void 0 : _a.map(cp => cp.Prefix);
                if (prefixes != null && prefixes.length > 0) {
                    dirs = prefixes;
                }
            }
            const nextMarker = awsResult.NextMarker;
            return { files, dirs, nextMarker };
        }
        catch (ex) {
            throw ex;
        }
    }
    async copyCloudFile(cf, dest) {
        if (dest.bucket.type !== this.type) {
            throw new Error(`destBucket type ${dest.bucket.type} does not match source bucket type ${this.type}. For now, cross bucket type copy not supported.`);
        }
        const sourcePath = cf.Key;
        const params = {
            CopySource: `${this.name}/${sourcePath}`,
            Bucket: dest.bucket.name,
            Key: dest.path
        };
        await this.s3.copyObject(params).promise();
    }
    async downloadCloudFile(rawFile, localPath) {
        const remotePath = rawFile.Key;
        const params = { ...this.baseParams, ...{ Key: remotePath } };
        const remoteReadStream = this.s3.getObject(params).createReadStream();
        const localWriteStream = fs_extra_plus_1.createWriteStream(localPath);
        const writePromise = new Promise((resolve, reject) => {
            localWriteStream.once('close', () => {
                resolve();
            });
            localWriteStream.once('error', (ex) => {
                reject(ex);
            });
            remoteReadStream.pipe(localWriteStream);
        });
        await writePromise;
    }
    async uploadCloudFile(localPath, remoteFilePath, contentType) {
        const readable = fs_extra_plus_1.createReadStream(localPath);
        const awsResult = await this.s3.putObject({ ...this.baseParams, ...{ Key: remoteFilePath, Body: readable, ContentType: contentType } }).promise();
        // TODO: probably check the awsResult that match remoteFilePath
        return { Key: remoteFilePath };
    }
    async downloadAsText(path) {
        const params = { ...this.baseParams, ...{ Key: path } };
        const obj = await this.s3.getObject(params).promise();
        const content = obj.Body.toString();
        return content;
    }
    async uploadCloudContent(path, content, contentType) {
        await this.s3.putObject({ ...this.baseParams, ...{ Key: path, Body: content, ContentType: contentType } }).promise();
    }
    async createReadStream(path) {
        const params = { ...this.baseParams, ...{ Key: path } };
        const obj = this.s3.getObject(params);
        if (!obj) {
            throw new Error(`Object not found for ${path}`);
        }
        return obj.createReadStream();
    }
    async createWriteStream(path, contentType) {
        const writable = new S3UploadWriteStream();
        const params = { ...this.baseParams, ...{ Key: path, ContentType: contentType }, Body: writable };
        const uploadCtrl = this.s3.upload(params);
        // NOTE: We use the S3UploadWriteStream trigger finish and close stream even when the upload is done
        uploadCtrl.promise().then(() => {
            writable.triggerFinishAndClose();
        });
        return writable;
    }
    async deleteCloudFile(path) {
        // NOTE: For aws API, the s3.deleteObject seems to return exactly the same if the object existed or not. 
        //       Therefore, we need to do an additional ping to know if the file exist  or not to return true/false
        const exists = await this.exists(path);
        if (exists) {
            // NOTE: between the first test and this delete, the object might have been deleted, but since s3.deleteObjecct
            //       does not seems to tell if the object exits or not, this is the best can do.
            await this.s3.deleteObject({ ...this.baseParams, ...{ Key: path } }).promise();
            return true;
        }
        else {
            process.stdout.write(` - Skipped (object not found)\n`);
            return false;
        }
    }
}
exports.S3Driver = S3Driver;
