"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getGsDriver = void 0;
const storage_1 = require("@google-cloud/storage");
const stream_1 = require("stream");
const micromatch = require("micromatch");
async function getGsDriver(cfg) {
    // TODO: valid cfg
    const googleStorageConf = {
        projectId: cfg.project_id,
        credentials: {
            client_email: cfg.client_email,
            private_key: cfg.private_key
        }
    };
    const storage = new storage_1.Storage(googleStorageConf);
    const googleBucket = storage.bucket(cfg.bucketName);
    return new GcpDriver(googleBucket);
}
exports.getGsDriver = getGsDriver;
class GcpDriver {
    constructor(googleBucket) {
        this.googleBucket = googleBucket;
    }
    get type() {
        return 'gs';
    }
    get name() {
        return this.googleBucket.name;
    }
    toFile(googleFile) {
        if (!googleFile) {
            throw new Error(`No googleFile`);
        }
        const size = (googleFile.metadata.size) ? Number(googleFile.metadata.size) : undefined;
        return {
            path: googleFile.name,
            size,
            updated: googleFile.metadata.updated,
            contentType: googleFile.metadata.contentType
        };
    }
    getPath(obj) {
        return obj.name;
    }
    async exists(path) {
        // Note: note gcp has specific file.exists method
        const result = await this.googleBucket.file(path).exists();
        return result[0];
    }
    async getCloudFile(path) {
        const googleFile = this.googleBucket.file(path);
        try {
            return (await googleFile.get())[0];
        }
        catch (ex) {
            // not found return null, as per getFile design.
            if (ex.code === 404) {
                return null;
            }
            // otherwise, propagate exception 
            else {
                throw ex;
            }
        }
    }
    /**
     *
     * @param path prefix path or glob (the string before the first '*' will be used as prefix)
     */
    async listCloudFiles(opts) {
        const { prefix, glob, directory, limit, marker } = opts;
        // build the query options and perform the re	quest
        let baseQuery = { autoPaginate: false };
        if (directory === true) {
            baseQuery.delimiter = '/';
        }
        if (limit != null) {
            baseQuery.maxResults = limit;
        }
        if (marker != null) {
            baseQuery.pageToken = marker;
        }
        let getListOpts = (prefix) ? { ...baseQuery, prefix } : baseQuery;
        // Note: need to fix the google bucket getFiles return type ([1] can be any or null)
        const [gfList, info, meta] = await this.googleBucket.getFiles(getListOpts);
        // get the dirs
        const dirs = (directory && meta.prefixes) ? meta.prefixes : undefined;
        // get pageToken (nextMarker)
        const nextMarker = info === null || info === void 0 ? void 0 : info.pageToken;
        // if glob, filter the data further
        let files = (!glob) ? gfList : gfList.filter(gf => micromatch.isMatch(gf.name, glob));
        return { files, dirs, nextMarker };
    }
    async downloadCloudFile(cf, localPath) {
        await cf.download({ destination: localPath });
    }
    async uploadCloudFile(localFilePath, remoteFilePath, contentType) {
        const googleBucket = this.googleBucket;
        const googleFile = (await googleBucket.upload(localFilePath, { destination: remoteFilePath, contentType }))[0];
        return googleFile;
    }
    async copyCloudFile(cf, dest) {
        if (dest.bucket.googleBucket == null) {
            throw new Error(`destBucket type ${dest.bucket.type} does not match source bucket type ${this.type}. For now, cross bucket type copy not supported.`);
        }
        const destGoogleBucket = dest.bucket.googleBucket;
        const destFile = destGoogleBucket.file(dest.path);
        await cf.copy(destFile);
    }
    async deleteCloudFile(path) {
        const googleFile = this.googleBucket.file(path);
        if (googleFile) {
            try {
                await googleFile.delete();
            }
            catch (ex) {
                // if not found, just return false.
                if (ex.code === 404) {
                    process.stdout.write(` - Skipped (object not found)\n`);
                    return false;
                }
                else {
                    process.stdout.write(` - FAILED - ABORT - Cause ${ex}\n`);
                    throw ex;
                }
            }
            // TODO: Probably needs to return true only if deleted. 
            return true;
        }
        else {
            return false;
        }
    }
    async downloadAsText(path) {
        const googleFile = this.googleBucket.file(path);
        const buffer = await googleFile.download();
        return buffer.toString();
    }
    async uploadCloudContent(path, content, contentType) {
        const googleFile = this.googleBucket.file(path);
        const uploadReadable = new stream_1.Readable();
        return new Promise(function (resolve, reject) {
            uploadReadable
                .pipe(googleFile.createWriteStream({ contentType }))
                .on('error', function (err) {
                reject(err);
            })
                .on('finish', function () {
                resolve();
            });
            uploadReadable.push(content);
            uploadReadable.push(null);
        });
    }
    async createReadStream(path) {
        const googleFile = this.googleBucket.file(path);
        return googleFile.createReadStream();
    }
    async createWriteStream(path, contentType) {
        const googleFile = this.googleBucket.file(path);
        return googleFile.createWriteStream({ contentType });
    }
}
